//----状態量と入力の変数-----
volatile float theta=0.0; //車輪角度
volatile float phai=0.0; //車体角度
volatile float dtheta=0.0; //車輪角速度
volatile float dphai=0.0; //車体角速度
volatile float u=0.0; //制御入力

volatile float theta_old=0.0;
volatile float dtheta_old=0.0;
volatile float phai_old=0.0;

//---フィードバックゲイン----
volatile float F[4]= { -1. , -299.43345924,   -1.75290489,  -50.14991414}; 

//---メインループ時間計測の変数----
volatile float dt ;
volatile unsigned long Timer=0.0;

//--センサー値格納用変数---
#include<Wire.h>
#define ACC_CONV 1671.8 //acceleration converter
#define GYRO_CONV 131.1 //deg/s converter
volatile uint8_t data_ag[14]; //to accomodate data
volatile float ay=0; 
volatile float az=0;
volatile float roll=0;
volatile float Lroll=0;
volatile float Lroll_old=0;
volatile float phai_OFFSET=15.00;
volatile float dphai_OFFSET=1.25;

//---エンコーダー読み取り用変数----
#define PWMA 10 // Motor PWMB
#define AIN2 9 // Motor Rotation
#define AIN1 8 // Motor Rotatio
#define STBY 7 // Motor Standby
#define BIN1 6 // Motor Rotation
#define BIN2 4 // Motor Rotation
#define PWMB 5 // Motor PWMA
#define PIN_ENCA 2 // Encoder
#define PIN_ENCB 3 // Encoder

volatile int valA,valB;
volatile int stateA,stateB;
volatile int edgeA,edgeB;
const float threshold_ON=1;
const float threshold_OFF=0;
volatile int value = 0; 

//---位置決め系用変数----- 
volatile float ang_ref=0.0; //位置決め系目標角度
volatile float dang_ref=0.0;
volatile float error =0; //偏差
volatile float error_itg=0;
#include <MsTimer2.h>
const float VREF = 5.0; 
const int PWM_MAX = 255; // PWM 100%
const int PWM_MIN = 0; // PWM 0%
volatile float uref=0.0;
volatile float U = 0.0;


void setup() {
  Serial.begin(115200); //シリアル通信開始
  
  //---エンコーダー用ピン設定---
  pinMode(AIN1,OUTPUT);
  pinMode(AIN2,OUTPUT);
  pinMode(BIN1,OUTPUT);
  pinMode(BIN2,OUTPUT);
  pinMode(PIN_ENCA, INPUT);
  pinMode(PIN_ENCB, INPUT);
  digitalWrite(PIN_ENCA, HIGH);
  digitalWrite(PIN_ENCB, HIGH);
  attachInterrupt(0, updateEncoder, CHANGE); 
  attachInterrupt(1, updateEncoder, CHANGE);
  
  //----加速度・ジャイロセンサーと通信----
  Wire.begin(); //I2C通信開始
  //加速度・ジャイロセンサーのレジスタを指定し通信開始
  i2cWriteReg(0x68,0x6b,0x00); //Turn on a sensor
  i2cWriteReg(0x68,0x1b,0x00); //-250 < deg/s <250
  i2cWriteReg(0x68,0x1c,0x00); // -2g < acc <2g
  
  //---位置決め系を1msごとに呼び出し設定----
  MsTimer2::set(1.0, locate);     
  MsTimer2::start();             
}

//I2Cレジスタ指定の関数
void i2cWriteReg(uint8_t ad, uint8_t reg, volatile uint8_t data ){

  Wire.beginTransmission(ad); //starts transmission
  Wire.write(reg); //select reg
  Wire.write(data); //write data
  Wire.endTransmission(); //ends transmission

}
//エンコーダー読み取り関数（４逓灰の原理）
void updateEncoder(){
  
  valA=digitalRead(PIN_ENCA);
  valB=digitalRead(PIN_ENCB);
  edgeA=0;
  
  if((valA==threshold_ON)&&(stateA==0)){
    stateA=1; edgeA=1;
  } //rising

  if((valA==threshold_OFF)&&(stateA==1)){
    stateA=0, edgeA=-1;
  } //faling
  edgeB=0;

  if((valB==threshold_ON)&&(stateB==0)){
    stateB=1; edgeB=1;
  } //rising

  if((valB==threshold_OFF)&&(stateB==1)){
    stateB=0; edgeB=-1;
  } //faling

  if (edgeA==1){
    if(stateB==0){
      value++;
    }
    else{
      value--;
    }
  }
  
  if (edgeA==-1){
    if(stateB==1){
      value++;
    }
    else{
      value--;
    }
  }
  
   if (edgeB==1){
    if(stateA==1){
      value++;
    }
    else{
      value--;
    }
  }

  if (edgeB==-1){
    if(stateA==0){
      value++;
    }
    else{
      value--;
    }
  }
  
}
//位置決め系(PID制御の目標角度追従)
void locate(){
  
  //----目標角度を受け取って，モーターへの指令電圧を決定する----
  error=ang_ref-theta;
  error_itg += 0.001 * error;
  uref=0.4*error + 0.030*error_itg -0.005*dtheta; //PIDゲイン
  U=uref*(PWM_MAX/VREF);

   int mag = constrain(abs(U), PWM_MIN, PWM_MAX);  //指令電圧決定
   digitalWrite(STBY, HIGH);
   //---モーターよ動け!-------
   if(U >= 0){
        analogWrite(PWMB, mag);
        digitalWrite(BIN1, HIGH);
        digitalWrite(BIN2, LOW);
        analogWrite(PWMA, mag);
        digitalWrite(AIN1, HIGH);
        digitalWrite(AIN2, LOW);
      
        }
        
   else{
      analogWrite(PWMB, mag);
      digitalWrite(BIN1, LOW);
      digitalWrite(BIN2, HIGH);
      analogWrite(PWMA, mag);
      digitalWrite(AIN1, LOW);
      digitalWrite(AIN2, HIGH);
        }
    //---モーター動きました------
    

         
}

//メインループ
void loop() {
  //-----ループ1回にかかる時間を計測---- 
  dt=(millis()-Timer)*0.001;
  Timer=millis();
 
  //-----ここが制御のメイン!-------------
  u= -( F[0]*theta+ F[1]*(phai)+ F[2]*dtheta+F[3]*dphai); //制御入力(状態フィードバック)
  
  dang_ref += u*dt;
  ang_ref += dang_ref*dt; //位置決め系への目標角度
  //-------メインおわり--------------
  
  //----エンコーダーの値を変換して車輪角度を更新!---
  
  int v;　　　　　　　//一時的な2バイト変数を作る
  noInterrupts();   //いったん割込禁止
  v = value;       //割り込み禁止状態でvalueをコピー
  interrupts();     //割込再開
　
  theta = -float(v) * 0.25;
  
  //----不完全微分で近似的に車輪角速度を更新!------
  dtheta= 80*(theta-theta_old) + 0.6703*dtheta_old;
  theta_old=theta;
  dtheta_old=dtheta;   
  //------車体角度と車体角速度を更新していく----------
  Wire.beginTransmission(0x68); //Starts transmission
  Wire.write(0x3b); //select data top
  Wire.endTransmission(); //ends transmission
  Wire.requestFrom(0x68,14); //get data from top to 14byte
  
  uint8_t i =0;
  while (Wire.available())
  {
    data_ag[i++]=Wire.read(); //accomodate data
  }
  
  //----ジャイロセンサーから車体角速度を更新!------
  dphai= (float)((data_ag[8] <<8 ) | data_ag[9])/GYRO_CONV+dphai_OFFSET;  //オフセットの補償
  
  //-------加速度センサーから車体角度を推定-----
  ay= (float)((data_ag[2] <<8 ) | data_ag[3])/ACC_CONV;
  az= (float)((data_ag[4] <<8 ) | data_ag[5])/ACC_CONV;
  if( millis()*0.001>40){
    phai_OFFSET=12;
  }
  roll=-(atan2(az,ay)*RAD_TO_DEG+ phai_OFFSET); //オフセットの補償
  Lroll=0.3297*roll + 0.6703*Lroll_old; //ローパスフィルター
  Lroll_old=Lroll;
  
  //-------推定した角度を相補フィルターにかけて車体角度を更新!---
  phai=0.999*(phai_old + dphai*dt) +0.001 * Lroll; //99.9%はジャイロの値を信じ，0.001%は加速度の値を信じる
  phai_old=phai;

  //---デバグに活躍するシリアルプロッタ関数---
  SERIAL_OUTPUT();
  
}

void SERIAL_OUTPUT(){
  //のぞいてみたい変数をここに書く
  Serial.print("phai: ");
  Serial.println(phai); 
}
